
Map = class("Map")

function Map:new(width, height)
	local m = instance(self)
	m.cells = {}
	m.waves = {}
	m.particles = {}
	m.creeps = {}
	m.lights = {}
	m.width = width
	m.height = height
	m.pulseTimer = 0
	return m
end

function Map:addCreep(creep)
	table.insert(self.creeps, creep)
	creep.map = self
	return creep
end

local function orderSort(a,b)
	return a.order < b.order
end

function Map:sortCreeps()
	shellsort(self.creeps, orderSort)
end

function Map:addLight(light)
	table.insert(self.lights, light)
	light.map = self
	return light
end


function Map:addCell(cell)
	table.insert(self.cells, cell)
	cell.map = self
	self:clampCell(cell)
	return cell
end

function Map:addWave(wave)
	table.insert(self.waves, wave)
	wave.map = self
end

function Map:addParticle(particle)
	table.insert(self.particles, particle)
	return particle
end

function Map:clampCell(cell)
	if cell.x-cell.radius < 0 then
		cell.x = cell.radius
		if cell.dx < 0 then
			cell.dx = cell.dx*-cell:getBounce()
		end
	elseif cell.x+cell.radius > self.width then
		cell.x = self.width - cell.radius
		if cell.dx > 0 then
			cell.dx = cell.dx*-cell:getBounce()
		end
	end

	if cell.y-cell.radius < 0 then
		cell.y = cell.radius
		if cell.dy < 0 then
			cell.dy = cell.dy*-cell:getBounce()
		end
	elseif cell.y+cell.radius > self.height then
		cell.y = self.height - cell.radius
		if cell.dy > 0 then
			cell.dy = cell.dy*-cell:getBounce()
		end
	end
end

function Map:wrap(ent, buffer)
	local buffer = buffer or 0
	if ent.x < -buffer then
		ent.x = ent.x + self.width + buffer*2
	elseif ent.x > self.width + buffer then
		ent.x = ent.x - self.width - buffer*2
	end
	if ent.y < -buffer then
		ent.y = ent.y + self.height + buffer*2
	elseif ent.y > self.height + buffer then
		ent.y = ent.y - self.height - buffer*2
	end
end

function Map:removeWaves()
	for i=1,#self.waves do
		table.remove(self.waves, 1)
	end
 end

function Map:update(time)
	for index, particle in pairs(self.particles) do
		particle:update(time)
	end
	for index, light in pairs(self.lights) do
		light:update(time)
	end
	for index, creep in pairs(self.creeps) do
		creep:update(time)

		self:wrap(creep, 256)
	end

	if self.pulseTimer > 0 then
		self.pulseTimer = math.max(0,self.pulseTimer - time)
	end

	table.removeCompletelyIf(self.particles)
	for index, wave in pairs(self.waves) do
		wave:update(time)
	end
	table.removeCompletelyIf(self.waves)
	
	local lastOwner = nil
	self.hasOpponents = false

	for index, cell in pairs(self.cells) do
		cell.owner.dead = false
		if lastOwner and lastOwner ~= cell.owner then
			self.hasOpponents = true
		else
			lastOwner = cell.owner
		end
		cell:update(time)
		cell:updateMotion(time)
		for index, bump in pairs(self.cells) do
			if bump ~= cell then
				local dist = math.distance(cell.x, cell.y, bump.x, bump.y)
				if dist < cell.radius*cell.fadeIn + bump.radius*bump.fadeIn then
					local ang = math.angleBetweenPoints(cell.x, cell.y, bump.x, bump.y)
					local factor = ((cell.radius + bump.radius) - dist)/(cell.radius + bump.radius)*4000*cell.fadeIn*bump.fadeIn
					if cell.spiked and cell.owner ~= bump.owner then
						if bump.spiked then
							factor = factor*3
						else
							bump:destroy()
						end
					end 
					cell:addSpeed(math.cos(ang+math.pi)*factor/cell:getWeight()*time, math.sin(ang+math.pi)*factor/cell:getWeight()*time)
					bump:addSpeed(math.cos(ang)*factor/bump:getWeight()*time, math.sin(ang)*factor/bump:getWeight()*time)
				end
				if bump.owner == cell.owner and cell.suckTimer > 0 then
					local ang = math.angleBetweenPoints(bump.x, bump.y, cell.x, cell.y)
					bump:addSpeed(math.cos(ang)*1000/(2+dist*0.7)*cell.suckTimer*time*bump.fadeIn, math.sin(ang)*1000/(2+dist*0.7)*cell.suckTimer*time*bump.fadeIn)
				end
			end
		end

		local dist = math.distance(cell.x, cell.y, self.width*0.5, self.height*0.5)
		if dist < 200 then
			local factor = (200-dist)/200*(0.2 + 0.8*self.pulseTimer)
			local ang = math.angleBetweenPoints(self.width*0.5, self.height*0.5, cell.x, cell.y)
			cell:addSpeed(math.cos(ang)*200*factor*time, math.sin(ang)*200*factor*time)
		end

		cell:applyFriction(time)

		self:clampCell(cell)
	end
	if not self.winner then
		if not self.hasOpponents and lastOwner then
			self.winner = lastOwner
			self.winner.wins = self.winner.wins +1
			states.get("music"):setNextMusic("win.ogg", 0.1, true)
			_config.musicVolume = 1
		end
	end
	local anyInFront = false
	for index, cell in pairs(self.cells) do
		for index, wave in pairs(self.waves) do
			
				local dist = math.distance(wave.x, wave.y, cell.x, cell.y)
				local closeDist = dist-cell.radius
				local longDist = dist+cell.radius
				if wave.radius <= longDist and wave.radius >= closeDist then
					local angTo = math.angleBetweenPoints(wave.x, wave.y, cell.x, cell.y)
					if wave:isAngleActive(angTo) then
						if wave.owner ~= cell.owner then
							if cell.fadeIn < 1 then
								cell.hitWanted = true
							else
								local factor = (cell.radius - math.abs(dist - wave.radius))/cell.radius
								if factor > cell.hitFactor then
									cell.lastWaveHit = wave
								end
								cell.hitFactor = math.max(cell.hitFactor, factor)
							end
						end
						--else
						--	cell.waveProtected = wave
						--end
					end
				end
				if wave.radius < longDist and wave.owner ~= cell.owner then
					anyInFront = true
				end
			--end
		end
	end
	self.fieldIsActive = anyInFront

	for index, cell in ipairs(self.cells) do
		if cell.fadeIn >= 1 then
			if not cell.owner.nextToDie then
				cell.owner.nextToDie = cell
			end
			if cell.lastHitFactor > 0 and cell.hitFactor <= 0 then
				cell:destroy()
				if cell.lastWaveHit and cell.lastWaveHit.owner then
					cell.lastWaveHit.owner.killFeedback = 1
					cell.lastWaveHit.owner.killed = cell.lastWaveHit.owner
				end
			end
		end
	end
	table.removeCompletelyIf(self.cells)
end

function Map:getWavesAffecting(cell)
	local waves = {}
	for index, wave in pairs(self.waves) do
		if wave.owner ~= cell.owner then
			local angTo = math.angleBetweenPoints(wave.x, wave.y, cell.x, cell.y)
			if wave:isAngleActive(angTo) then
				local dist = math.distance(wave.x, wave.y, cell.x, cell.y)
				local closeDist = dist-cell.radius
				local longDist = dist+cell.radius
				if wave.radius <= longDist and wave.radius >= closeDist then
					table.insert(waves, wave)
				end
			end
		end
	end
	return waves
end

function Map:bgRender()
	if states.get("menu") then
		local fade = math.min(1,states.get("menu").startTimer/4)
		local time = states.get("menu").time
		video.renderSpriteState(SPRITES.letter_c, window.w*0.2+ math.cos(time*0.5-0.5)*5, window.h*0.4 + math.cos(time*0.5-0.5)*15, 0.7*fade, math.cos(time*0.5-0.5)*math.pi*0.035, 100*fade, 255, 255, 255)
		video.renderSpriteState(SPRITES.letter_e, window.w*0.35+ math.cos(time*0.5-3.5)*7, window.h*0.4 + math.cos(time*0.45)*15, 0.7*fade,  math.cos(time*0.45-2.11)*math.pi*0.02, 100*fade, 255, 255, 255)
		video.renderSpriteState(SPRITES.letter_l, window.w*0.5+ math.cos(time*0.25-2.5)*5, window.h*0.4 + math.cos(time*0.55 + 1.5)*6, 0.7*fade,  math.cos(time*0.565-0.25)*math.pi*0.012, 100*fade, 255, 255, 255)
		video.renderSpriteState(SPRITES.letter_l, window.w*0.65+ math.cos(time*0.15-1.5)*7, window.h*0.4 + math.cos(time*0.52+0.4)*9, 0.7*fade,  math.cos(time*0.2125-0.25)*math.pi*0.032, 100*fade, 255, 255, 255)
		video.renderSpriteState(SPRITES.letter_s, window.w*0.8+ math.cos(time*0.75-0.5)*5, window.h*0.4 + math.cos(time*0.51+0.2)*15, 0.7*fade,  math.cos(time*0.75-0.5)*math.pi*0.015, 100*fade, 255, 255, 255)
	end

	for index, creep in pairs(self.creeps) do
		creep:render(creep.x, creep.y, 1)
	end

	for index, light in pairs(self.lights) do
		light:render(light.x, light.y, 1)
	end

	video.renderSpriteState(SPRITES.corner, 0,0,1, 0, 255, 							155,250,0)
	video.renderSpriteState(SPRITES.corner, self.width,0,1, 0, 255, 				155,250,0, true)
	video.renderSpriteState(SPRITES.corner, self.width,self.height,1, math.pi, 255, 155,250,0)
	video.renderSpriteState(SPRITES.corner, 0,self.height,1, math.pi, 255, 			155,250,0, true)
end

function Map:render()
	for index, wave in pairs(self.waves) do
		wave:render(wave.x, wave.y, 1)
	end
	for index, particle in pairs(self.particles) do
		particle:render(particle.x, particle.y, 1, 0, 255)
	end
	for index, cell in pairs(self.cells) do
		cell:render(cell.x, cell.y, 1, 0, 255)
	end
end

Wave = class("wave")

function Wave:new(owner, x, y, speed, radius)
	c = instance(self)
	c.x = x
	c.y = y
	c.owner = owner
	c.speed = speed
	c.radius = radius or 1
	c.disabled = {}
	c.segs = {}
	for i=0,63 do
		c.segs[i] = 1
	end
	c.fade = 1
	return c
end

function Wave:update(time)
	self.radius = self.radius + self.speed*time
	if self.radius > math.max(self.map.width,self.map.height)*1.5 then
		self:remove()
	end
	if self.broken then
		local color = self:getColor()
		for i=0,63 do
			if math.random() <= time*20 then
				local before = self.segs[(i-1)%64]
				local current = self.segs[i]
				local after = self.segs[(i+1)%64]
				local x,y = nil,nil
				if (before > 0 and current > 0 and after <= 0 ) then
					x,y = self:getSegPosition(i)
				elseif (before <= 0 and current > 0 and after > 0 ) then
					x,y = self:getSegPosition(i-1)
				end
				if x then
					local a = self.map:addParticle(Particle:new(x,y,math.random()*50-25,math.random()*50-25,2,255,color.r,color.g,color.b,0, color.r, color.g, color.b, 0.6, 15))
				end
			end
		end
	end
	self.fade = self.fade - time*5
end
function Wave:remove()
	self._remove = true
end

function Wave:getSegPosition(index)
	return self.x + math.cos(-(index+1)/64*math.pi*2)*self.radius,  self.y + math.sin(-(index+1)/64*math.pi*2)*self.radius
end

function Wave:render(x,y, scale, angle, a)
	local color = self.owner and self.owner.color or COLORS.white
	video.renderSpriteCircleLineBinaryMask(x,y, self.radius-6, 1,1, 64, 0, 255, color.r, color.g, color.b, math.min(self.radius,16), self.segs, SPRITES.wave)
	if self.fade > 0 then
		video.renderAdditiveSpriteState(SPRITES.shockwave, x, y, self.radius/256, self.fade*255, color.r, color.g, color.b)
	end
end	

function Wave:disableSegsAt(x,y,range)
	for index, seg in pairs(self.segs) do
		if seg > 0 then
			local angTo = math.angleBetweenPoints(self.x, self.y,x,y)
			local fromX, fromY = x-math.cos(angTo)*50, y-math.sin(angTo)*50
			local sx,sy = self.x + math.cos(-index/64*math.pi*2)*self.radius,  self.y + math.sin(-index/64*math.pi*2)*self.radius
			local dist = math.distance(x,y, sx,sy)
			if dist < range then
				self.segs[index] = 0
				local angTo = math.angleBetweenPoints(fromX, fromY,sx, sy)
				local color = self:getColor()
				for i=1,5 do
					local spawnAngle = -index/64*math.pi*2 + math.random()*math.pi*2/64
					local a = self.map:addParticle(Particle:new(self.x + math.cos(spawnAngle)*self.radius,self.y + math.sin(spawnAngle)*self.radius,math.cos(angTo)*200,math.sin(angTo)*200,2,255,color.r,color.g,color.b,0, color.r, color.g, color.b, 0.2, 15))
				end
				--local sx2,sy2 = self.x + math.cos(-(index+1)/64*math.pi*2)*self.radius,  self.y + math.sin(-(index+1)/64*math.pi*2)*self.radius
				--local angTo = math.angleBetweenPoints(fromX, fromY, sx2,sy2)
				--local b = self.map:addParticle(Particle:new(sx2,sy2,math.cos(angTo)*100,math.sin(angTo)*100,2,255,color.r,color.g,color.b,0, color.r, color.g, color.b, 0.2, 5))
				
				--a:addSpeed(math.cos(math.randomAngle())*math.random()*1500, math.cos(math.randomAngle())*math.random()*1500)
				--b:addSpeed(math.cos(math.randomAngle())*math.random()*1500, math.cos(math.randomAngle())*math.random()*1500)
				--a:setLine(b)
				self.broken = true
			end
		end
	end
end

function Wave:getColor()
	if self.owner then
		return self.owner.color
	else
		return COLORS.white
	end
end

function Wave:isAngleActive(angle)
	local index = math.round(-angle/math.pi/2*64)
	if index < 0 then
		index = index + 64
	end
	if index > 63 then
		index = index - 64
	end
	return self.segs[index] > 0 
end


Cell = class("Cell")

function Cell:new(owner, x, y, radius, dx, dy)
	c = instance(self)
	c.x = x
	c.y = y
	c.radius = radius or 1
	c.dx = dx or 0
	c.dy = dy or 0
	c.owner = owner
	c.fadeIn = 0
	c.hitFactor = 0
	c.lastHitFactor = 0
	c.hitFeedback = 0
	c.nuclea = 1
	c.time = 0
	c.suckTimer = 0
	c.nucX = 0
	c.nucY = 0
	return c
end

function Cell:remove()
	self._remove = true
end

function Cell:destroy()
	if not self._remove then
		self._remove = true
		for i=1,10 do
			local ang = math.randomAngle()
			local dist = math.random()*self.radius
			self.map:addParticle(Particle:new(self.x+ math.cos(ang)*dist,self.y + math.sin(ang)*dist,math.cos(ang)*dist*55,math.sin(ang)*dist*55,4,255,self.owner.color.r,self.owner.color.g,self.owner.color.b,0, self.owner.color.r*0.5,self.owner.color.g*0.5,self.owner.color.b*0.5, 0.25, 0.4))
		end
		if self.spiked then
			self:setNuclea(math.min(self.nuclea, 4))
			self:split(2, 3)
		end
		audio.playSound("sfx/pop"..math.random(1,10)..".ogg", 1, self:getPan(), 1)
	end
end

function Cell:suicide()
	if not self._remove then
		self._remove = true
		for i=1,10 do
			local ang = math.randomAngle()
			local dist = self.radius
			local speed = math.random()*self.radius
			self.map:addParticle(Particle:new(self.x+ math.cos(ang)*dist,self.y + math.sin(ang)*dist,math.cos(ang+math.pi)*speed,math.sin(ang+math.pi)*speed,3,255,self.owner.color.r,self.owner.color.g,self.owner.color.b,0, self.owner.color.r*0.5,self.owner.color.g*0.5,self.owner.color.b*0.5, 1.25, 2.4))
		end
		if self.spiked then
			self:setNuclea(math.min(self.nuclea, 4))
			self:split(2, 3)
		end
		audio.playSound("sfx/pop"..math.random(1,10)..".ogg", 1, self:getPan(), 1.4)
	end
end


function Cell:split(factor, amt)
	local amt = amt or 2
	local angle = math.randomAngle()
	local maxSpeed = 30*(factor or 1)
	for i=1,amt do
		local speed = math.random()*maxSpeed
		local nuclea = math.max(1,1 + (self.nuclea-1)/2)
		local c = self.map:addCell(Cell:new(self.owner, self.x, self.y, 17 + nuclea*2, self.dx + math.cos(angle+i*math.pi*2/amt)*speed, self.dy + math.sin(angle+i*math.pi*2/amt)*speed))
		c.nuclea = nuclea
		c.spitTimer = 1
	end
	self:remove()
end

function Cell:getPan()
	return (self.x-self.map.width*0.5)/self.map.width*0.8
end

function Cell:setNuclea(v)
	self.nuclea = v
	if self.nuclea >= 5 then
		self.spiked = true
	end
end

function Cell:update(time)
	if self.startFade then
		self.startFade = math.min(1,self.startFade + time*5)
		if self.startFade >= 1 then
			self.startFade = nil
		end
	end
	if self.splitTimer then
		self.splitTimer = self.splitTimer - time*2
		if self.splitTimer <= 0 then
			self.splitTimer = nil
		end
	end
	if self.fadeIn < 1 then
		if self.hitWanted then
			self.fadeIn = math.min(0.9,self.fadeIn + time/2)
		else
			self.fadeIn = math.min(1,self.fadeIn + time/2)
		end
	end
	self.lastHitFactor = self.hitFactor
	self.hitFactor = 0
	self.hitWanted = false
	self.waveProtected = false
	if self.hitFeedback > 0 then
		self.hitFeedback = math.max(0,self.hitFeedback - time*7)
	end
	if self.suckTimer > 0 then
		self.suckTimer = self.suckTimer - time
	end
	self.time = self.time + time


	self.nucX = math.approach(self.nucX + (math.random() - 0.5)*2*time, 0, time + math.abs(self.nucX)*time)
	self.nucY = math.approach(self.nucY + (math.random() - 0.5)*2*time, 0, time + math.abs(self.nucY)*time)
end	

function Cell:applyFriction(time)
	local speed = math.distance(0,0,self.dx, self.dy)
	self.dx = math.approach(self.dx, 0, speed*time*0.2)
	self.dy = math.approach(self.dy, 0, speed*time*0.2)
end

function Cell:updateMotion(time)
	self.x = self.x + self.dx*time
	self.y = self.y + self.dy*time
end

function Cell:addSpeed(dx, dy)
	self.dx = self.dx + dx
	self.dy = self.dy + dy
end

function Cell:getBounce()
	return 0.2
end

function Cell:getWeight()
	return self.radius*self.radius
end	

function Cell:render(x,y,scale,angle,a)
	if self.startFade then
		a = math.min(255,a*(1-self.startFade))
	end
	local bright = 1
	if self.owner.nextToDie == self then
		bright = 0.8 - math.max(0,math.cos(self.time*2)*0.1)
	end
	local colorFade = 0.3 + 0.4*self.owner.killFeedback*self.owner.killFeedback--(0.33+ 0.67*self.fadeIn*self.fadeIn) + 0.5*self.hitFeedback
	local glowScale = 1
	local scale = scale * 0.95 + scale*self.fadeIn*self.fadeIn*self.fadeIn*0.05+ 0.5*self.hitFeedback
	a = math.min(255,a*0.2 + a*0.8*self.fadeIn)
	if self.hitFactor > 0 then
		colorFade = math.max(0,colorFade - self.hitFactor*0.3)
		glowScale = glowScale + 0.2 + self.hitFactor*0.75 
	end
	if self.fadeIn < 1 then
		a = math.min(128 + 128*math.max(0,(self.fadeIn-0.93))/0.07,a)
	end
	glowScale = glowScale + 0.33*self.owner.killFeedback*self.owner.killFeedback
	local invColorFade = 1-colorFade
	video.renderSpriteState(SPRITES.cellBodyGlow, x, y, self.radius/29*scale*glowScale*self.fadeIn, 0, a, 255*invColorFade + colorFade*self.owner.color.r, 255*invColorFade + colorFade*self.owner.color.g, 255*invColorFade +colorFade*self.owner.color.b)
	local vScale, hScale = 1*(1-self.hitFactor*0.3) + 0.3*self.hitFactor*math.cos(38*self.time), 1*(1-self.hitFactor*0.3) + 0.3*self.hitFactor*math.sin(38*self.time)
	if self.spiked then
		for i=0,8 do
			local an = math.pi*2/9*i
			video.renderSpriteState(SPRITES.cellSpike, x + math.cos(an)*self.radius*1*hScale, y + math.sin(an)*self.radius*1*vScale, self.radius/29, an-math.pi*0.5,a, 255*invColorFade + colorFade*self.owner.color.r, 255*invColorFade + colorFade*self.owner.color.g, 255*invColorFade +colorFade*self.owner.color.b)
		end
	end

	local scaledRad = self.radius*scale
	local x1,y1 = x - scaledRad*hScale, y - scaledRad*vScale
	local x2,y2 = x + scaledRad*hScale, y - scaledRad*vScale
	local x3,y3 = x - scaledRad*hScale, y + scaledRad*vScale
	local x4,y4 = x + scaledRad*hScale, y + scaledRad*vScale
	local cellColorFade = 1 - math.min(1,math.abs(math.cos(68*self.time))*self.hitFactor*2)
	local invCellColorFade = 1-cellColorFade
	local glowColor = self.owner.color
	if self.owner.killFeedback > 0 then
		glowColor = self.owner.killed.color
	end
	local spr = self.spiked and SPRITES.cellBodySpiked or SPRITES.cellBody
	video.renderSpriteStateFreeShape(spr, x1,y1,x2,y2,x3,y3,x4,y4, math.min(255,a), 255*invCellColorFade + cellColorFade*self.owner.color.r*bright, 255*invCellColorFade + cellColorFade*self.owner.color.g*bright, 255*invCellColorFade +cellColorFade*self.owner.color.b*bright)
	local nucleaFade = 0.8 - 0.2*self.owner.killFeedback*self.owner.killFeedback
	local nucleaScale = 1 + 0.3*self.hitFactor
	if self.hitFactor >= 0.75 - (self.nuclea-1)*0.04 then
		nucleaFade = nucleaFade - 0.2 - self.hitFactor*0.25
	end
	local invNucleaFade = 1-nucleaFade
	local nucleaAlpha = math.max(0,(self.fadeIn-0.95))/0.05
	
	video.renderSpriteState(SPRITES.cellNucleus, x + self.nucX, y + self.nucY, self.nuclea/2*scale*(1+self.hitFactor*0.4)*nucleaScale, 0, math.min(255,nucleaAlpha*a), 0*invNucleaFade + nucleaFade*self.owner.color.r, 0*invNucleaFade + nucleaFade*self.owner.color.g, 0*invNucleaFade +nucleaFade*self.owner.color.b)
	local angTo = math.angleBetweenPoints(x,y, window.w*0.4, -window.h*0.7)
	video.renderSpriteState(SPRITES.cellHighlight, x + math.cos(angTo)*self.radius*0.7*self.fadeIn,  y + math.sin(angTo)*self.radius*0.7*self.fadeIn, self.fadeIn*self.radius/25, math.min(255,nucleaAlpha*a), 255, 255, 255)
	if self.fadeIn < 1 then
		local fad = 1-self.fadeIn*self.fadeIn*self.fadeIn*0.9
		local fadAlpah = math.min(self.fadeIn*0.7,1)
		local sc = self.fadeIn*self.fadeIn*self.fadeIn*0.65
		video.renderSpriteState(SPRITES.cellNucleus, x + self.nucX + self.radius*0.7*fad, y + self.nucY, self.nuclea/2*scale*nucleaScale*sc, 0, math.min(255,a*fadAlpah), 0*invNucleaFade + nucleaFade*self.owner.color.r, 0*invNucleaFade + nucleaFade*self.owner.color.g, 0*invNucleaFade +nucleaFade*self.owner.color.b)
		video.renderSpriteState(SPRITES.cellNucleus, x + self.nucX - self.radius*0.7*fad, y + self.nucY, self.nuclea/2*scale*nucleaScale*sc, 0, math.min(255,a*fadAlpah), 0*invNucleaFade + nucleaFade*self.owner.color.r, 0*invNucleaFade + nucleaFade*self.owner.color.g, 0*invNucleaFade +nucleaFade*self.owner.color.b)
		video.renderSpriteState(SPRITES.cellNucleus, x + self.nucX, y + self.nucY+ self.radius*0.7*fad, self.nuclea/2*scale*nucleaScale*sc, 0, math.min(255,a*fadAlpah), 0*invNucleaFade + nucleaFade*self.owner.color.r, 0*invNucleaFade + nucleaFade*self.owner.color.g, 0*invNucleaFade +nucleaFade*self.owner.color.b)
		video.renderSpriteState(SPRITES.cellNucleus, x + self.nucX, y + self.nucY- self.radius*0.7*fad, self.nuclea/2*scale*nucleaScale*sc, 0, math.min(255,a*fadAlpah), 0*invNucleaFade + nucleaFade*self.owner.color.r, 0*invNucleaFade + nucleaFade*self.owner.color.g, 0*invNucleaFade +nucleaFade*self.owner.color.b)
	end
	if self.splitTimer and self.splitTimer > 0 then
		video.renderAdditiveSpriteState(SPRITES.shockwave, x, y, self.radius/256, self.splitTimer*255, self.owner.color.r, self.owner.color.g, self.owner.color.b)
	end
end

Particle = class("Particle")

function Particle:new(x,y,dx,dy,radius,a,r,g,b,enda, endr, endg, endb, lifetime, friction)
	local p = instance(self)
	p.x = x
	p.y = y
	p.dx = dx or 0
	p.dy = dy or 0
	p.radius = radius or 1
	p.startColor = Color:new(a,r,g,b)
	p.endColor = Color:new(enda, endr, endg, endb)
	p.a = a or 255
	p.r = r or 255
	p.g = g or 255
	p.b = b or 255
	p.friction = friction or 0
	p.lifetime = lifetime or 1
	p.time = 0
	return p
end

function Particle:update(time)
	self.time = self.time + time
	self.x = self.x + self.dx*time
	self.y = self.y + self.dy*time

	local speed = math.distance(0,0,self.dx, self.dy)
	self.dx = math.approach(self.dx, 0, self.friction*time)
	self.dy = math.approach(self.dy, 0, self.friction*time)

	local prg = self.time/self.lifetime

	self.a = self.startColor.a*(1-prg) + self.endColor.a*prg
	self.r = self.startColor.r*(1-prg) + self.endColor.r*prg
	self.g = self.startColor.g*(1-prg) + self.endColor.g*prg
	self.b = self.startColor.b*(1-prg) + self.endColor.b*prg
	
	if self.lifetime <= self.time then
		self:remove()
	end
end

function Particle:addSpeed(dx, dy)
	self.dx = self.dx + dx
	self.dy = self.dy + dy
end

function Particle:remove()
	self._remove = true
end

function Particle:render(x,y, scale)
	if self.lineTo then
		video.renderSpriteLine(x,y,self.lineTo.x,self.lineTo.y, self.a, self.r, self.g, self.b, nil, nil, width, 1)
	else
		video.renderRectangle(x-self.radius*scale, y-self.radius*scale, self.radius*scale*2, self.radius*scale*2, self.a, self.r, self.g, self.b)
	end
end

function Particle:setLine(t)
	self.lineTo = t
end

CREEP_TYPES = {}

CREEP_TYPES.bigThing = {
	init = function(self)
		self.var = math.random(1,6)
	end,
	render = function(self, x, y)
		video.renderSpriteState(SPRITES.bigThings[self.var], x, y, self.scale, self.angle, 255*self.opacity, 248, 255, 235)
	end,
	diveChance = 0.05,
	diveAmount = 0.8,
	turnChance = 0.2,
	turnAmount = 0.20,
	speedChance = 1.4,
	speedAmount = 31.2,
	friction = 7.21,
}

CREEP_TYPES.particle = {
	init = function(self)
		self.var = math.random(1,10)
		self.brightness = math.random(1,2)*150-50
	end,
	render = function(self, x, y)
		video.renderSpriteState(SPRITES.particles[self.var], x, y, self.scale, self.angle, 90*self.opacity, self.brightness, self.brightness, self.brightness)
	end,
	diveChance = 0.15,
	diveAmount = 3,
	turnChance = 1.1,
	turnAmount = 0.1,
	speedChance = 1.5,
	speedAmount = 13.2,
	friction = 7.7,
}

CREEP_TYPES.creep = {
	init = function(self)
		self.var = math.random(1,9)
		local w,h = video.getSpriteStateSize(SPRITES.creep[self.var])
		self.wiggles = {}
		for i=1,math.random(4) do
			table.insert(self.wiggles, {var = math.random(1,7), w = (math.random()*w-w*0.5)*0.33, h = (math.random()*h-h*0.5)*0.33 })
		end
	end,
	render = function(self, x, y)
		local angle = self.angle - math.pi*0.5
		video.renderSpriteState(SPRITES.creep[self.var], x, y, self.scale, angle, 170*self.opacity, 240, 255, 160)
		for index, wiggle in pairs(self.wiggles) do
			video.renderSpriteState(SPRITES.wiggle[wiggle.var], x + math.cos(angle)*wiggle.w + math.cos(angle)*wiggle.h, y+ math.sin(angle)*wiggle.w + math.sin(angle)*wiggle.h, self.scale*0.75, self.speedAngle +math.cos(self.wiggling*1 + index)*math.pi*0.2, 100*self.opacity, 240, 255, 150)

			--+ 
		end
	end,
	diveChance = 0.05,
	diveAmount = 1,
	turnChance = 1.1,
	turnAmount = 1.1,
	speedChance = 0.5,
	speedAmount = 72.2,
	friction = 14.4,
}

CREEP_TYPES.snake = {
	init = function(self)
	end,
	render = function(self, x, y)
		video.renderSpriteState(SPRITES.snakeBody, x, y, self.scale, self.angle, 255*self.opacity, 240, 255, 150)
	end,
	diveChance = 0.05,
	diveAmount = 1,
	turnChance = 1.1,
	turnAmount = 1.1,
	speedChance = 2.5,
	speedAmount = 8.2,
	friction = 17.7,
}

Creep = class("Creep")

function Creep:def()
	return CREEP_TYPES[self.type]
end

function Creep:new(mainType, x,y,dx,dy,angle,scale, order)
	local p = instance(self)
	p.type = mainType
	p.x = x
	p.y = y
	p.dx = dx or 0
	p.dy = dy or 0
	p.angle = angle or 0
	p.scale = scale or 0
	p.order = order or 1
	p:def().init(p)

	p.turn = 0
	p.throttle = 0
	p.speed = 0
	p.rudder = 0
	p.time = 0
	p.speedAngle = 0
	p.wiggling = 0
	p.dive = 0
	p.depth = 0
	p.opacity = 1
	return p
end

function Creep:update(time)

	self.time = self.time + time
	self.x = self.x + self.dx*time
	self.y = self.y + self.dy*time

	local speed = math.distance(0,0,self.dx, self.dy)
	self.dx = math.approach(self.dx, 0, self:def().friction*time)
	self.dy = math.approach(self.dy, 0, self:def().friction*time)

	if math.random() < time*self:def().turnChance then
		self.turn = math.random()*self:def().turnAmount - math.random()*self:def().turnAmount
	end
	if math.random() < time*self:def().speedChance then
		self.throttle = math.max(0,math.random()*self:def().speedAmount - math.random()*self:def().speedAmount)
	end

	if math.random() < time*self:def().diveChance then
		self.dive = math.random()*self:def().diveAmount - math.random()*self:def().diveAmount
	end

	self.rudder = math.approach(self.rudder, self.turn, time*math.abs(self.rudder-self.turn)*0.5)
	self.speed = math.approach(self.speed, self.throttle, time*math.abs(self.throttle-self.speed)*0.5)

	self:addSpeed(math.cos(self.angle)*self.speed*time, math.sin(self.angle)*self.speed*time)
	self.angle = self.angle + self.rudder * time

	self.depth = self.depth + self.dive * time

	self.depth = math.approach(self.depth, 0, math.abs(self.depth)*0.1*time)

	self.dive = math.approach(self.dive, 0, math.abs(self.dive)*time/4)
	self.turn = math.approach(self.turn, 0, math.abs(self.turn)*time/4)
	self.throttle = math.approach(self.throttle, 0, math.abs(self.throttle)*time/4)

	local spAng = math.angleBetweenPoints(0,0,self.dx, self.dy)
	self.speedAngle = math.approachAngle(self.speedAngle, spAng, math.abs(math.pi*math.angleDifference(self.speedAngle, spAng))*time/40)
	self.wiggling = self.wiggling + self.speed*time + self.rudder*time

	self.scale = 1 + self.depth/10
	self.opacity = math.max(0,math.min(1,1 - self.depth/2))
end

function Creep:addSpeed(dx, dy)
	self.dx = self.dx + dx
	self.dy = self.dy + dy
end

function Creep:render(x,y)
	self:def().render(self, x, y)
end

Light = class("Light")

function Light:def()
	return CREEP_TYPES[self.type]
end

function Light:new(x,y)
	local p = instance(self)
	p.startX = x
	p.startY = y
	p.x = x
	p.y = y
	p.alpha = 50
	p.time = math.random()*1000000
	return p
end

function Light:update(time)
	self.x = self.startX + math.cos(self.time*0.2)*self.map.width*0.1 + math.sin(self.time*0.2111)*self.map.width*0.1 + math.cos(self.time*0.05211)*self.map.width*0.1 + math.cos(self.time*0.01333)*self.map.width*0.1
	self.y = self.startY + math.cos(self.time*0.223)*self.map.height*0.1 + math.sin(self.time*0.223111)*self.map.height*0.1 + math.cos(self.time*0.04211)*self.map.height*0.1 + math.cos(self.time*0.01137)*self.map.height*0.1
	self.alpha = 105 + math.cos(self.time*0.7)*32 + math.cos(self.time*1.8)*26  + math.cos(self.time*2.8111)*10
	self.time = self.time + time*2
end

function Light:render(x,y)
	video.renderAdditiveSpriteState(SPRITES.spotlight, x, y, 2, 1,  self.alpha,  self.alpha,  self.alpha)
end
